use crate::{cards::card::*, constants, utils};
use amadeus_lib::actions;
use amadeus_macro::either;
use lkt_lib::{Kara, LektorType};
use std::{borrow::Borrow, cmp::min, marker::PhantomData};

pub type KaraCardCollection<'a> = CardCollection<'a, KaraCard, Kara>;

/// Struct to hold a collection of cards
pub struct CardCollection<'a, CardType, LktType>
where
    CardType: Card<'a, LktType>,
    LktType: LektorType<'a>,
{
    name: String,
    content: Vec<CardType>,
    actions: Vec<actions::Action>,
    max_content: Option<usize>,
    phantom: PhantomData<&'a LktType>,
}

impl<'a, CardType: Card<'a, LktType>, LktType: LektorType<'a>> ToString
    for CardCollection<'a, CardType, LktType>
{
    fn to_string(&self) -> String {
        format!(
            "Collection {}: {}",
            self.name,
            self.content
                .iter()
                .map(|card| card.to_string())
                .fold(String::new(), |a, b| a + ", " + b.borrow())
        )
    }
}

impl<'a, CardType: Card<'a, LktType>, LktType: LektorType<'a>>
    CardCollection<'a, CardType, LktType>
{
    pub fn new(name: String) -> Self {
        Self {
            name,
            max_content: None,
            content: vec![],
            actions: vec![],
            phantom: PhantomData,
        }
    }

    pub fn with_max_content(&mut self, max_content: usize) -> &mut Self {
        self.max_content = either!(max_content != 0 => Some(max_content); None);
        self
    }

    pub fn add_card(&mut self, card: CardType) -> &mut Self {
        self.content.push(card);
        self
    }

    pub fn add_action(mut self, act: actions::Action) -> Self {
        self.actions.push(act);
        self
    }

    pub fn empty(&self) -> bool {
        self.content.is_empty()
    }

    pub fn fulfilled_actions(&mut self) -> Vec<(u64, actions::Action)> {
        let mut ret = Vec::new();
        for card in &mut self.content {
            let actions = card.fulfilled_actions();
            let actions_count = actions.len();
            if actions_count != 0 {
                let id = card.unique_id();
                ret.reserve(ret.len() + actions_count);
                for action in actions {
                    ret.push((id, action));
                }
            }
        }
        ret
    }

    pub fn render(&mut self, ui: &mut egui::Ui, dark_mode: bool) {
        ui.vertical(|ui| {
            ui.add_space(constants::PADDING);
            ui.horizontal(|ui| {
                ui.style_mut().override_text_style = Some(utils::font::heading1());
                ui.colored_label(constants::get_text_color(dark_mode), self.name.to_string());
                ui.style_mut().override_text_style = Some(utils::font::small_body());
                ui.label(format!(" ({} elements)", self.content.len()));
            });
            if CardType::NEED_HEADER_SEPARATOR {
                ui.add_space(constants::PADDING);
                ui.add(egui::Separator::default());
            }
            if self.empty() {
                ui.vertical_centered_justified(|ui| {
                    ui.add_space(constants::PADDING * 2.);
                    ui.add(egui::Spinner::new());
                    ui.allocate_space(ui.available_size())
                });
            } else {
                egui::ScrollArea::vertical()
                    .hscroll(false)
                    .always_show_scroll(false)
                    .max_width(f32::INFINITY)
                    .show(ui, |ui| {
                        ui.horizontal(|ui| ui.add_space(ui.available_width()));
                        let upper_bound = match self.max_content {
                            None => self.content.len(),
                            Some(x) => min(self.content.len(), x),
                        };
                        for card in &mut self.content[0..upper_bound] {
                            card.render(ui, dark_mode, &self.actions);
                            if CardType::NEED_SEPARATOR {
                                ui.add(egui::Separator::default());
                            }
                        }
                        if let Some(space) = CardType::BOTTOM_SPACE {
                            ui.add_space(space);
                        }
                    });
                ui.allocate_space(ui.available_size());
            }
        });
    }
}
